@@ -1,72 +1,85 @@
 import ast
 import json
 import logging
+import numbers
 import threading
 import uuid
 from inspect import signature
-from typing import Optional, TypeVar
+from typing import Optional, Any, Callable
 
 from pika import BlockingConnection, BasicProperties
 from pika.adapters.blocking_connection import BlockingChannel
 from pika.spec import Basic
 
-from snacks import APP
 from snacks.rabbit_config import RabbitConfig
 
-__all__ = ('Rabbit',)
-log = logging.getLogger(APP)
-T = TypeVar('T')
+__all__ = ('RabbitApp',)
+log = logging.getLogger('snacks')
 
 
-class Rabbit:
+# TODO snacks-django, finds all serializers, uses those.
+class RabbitApp:
     """A class to interface with RabbitMQ."""
 
     def __init__(self, rabbit_config: RabbitConfig) -> None:
         self.config = rabbit_config
+        log.info('RabbitApp configured.')
+        log.debug('host=%s', rabbit_config.host)
+        log.debug('port=%s', rabbit_config.port)
+        log.debug('exchange=%s', rabbit_config.exchange)
+        log.debug('virtual_host=%s', rabbit_config.virtual_host)
 
-    def publish(self, body: any, routing_key: str, **kwargs) -> None:
+    # TODO Make async?
+    def publish(self, body: Any, routing_key: str, **kwargs) -> None:
         """Publish message to a rabbit queue with the given routing key.
 
         :param body: The message to publish.
         :param kwargs: Keyword args to pass to pika publish.
         :param routing_key: The routing key.
         """
-        log.debug('Publishing key={} msg={}', routing_key, body)
+        log.debug('Publishing key=%s msg=%s', routing_key, body)
         mq_conn = BlockingConnection(self.config.params)
         channel: BlockingChannel = mq_conn.channel()
+        if 'to_json' in dir(body):
+            body = body.to_json()
+        elif isinstance(body, numbers.Number):
+            body = str(body)
         channel.basic_publish(
             exchange=self.config.exchange,
             routing_key=routing_key,
-            body=_get_body(body),
+            body=body,
             **kwargs
         )
 
-    def publish_and_receive(self,
-                            body: any,
-                            routing_key: str,
-                            response_type: T,
-                            time_limit: int = 10,
-                            headers: Optional[dict[str, any]] = None) -> T:
+    # TODO Make async?
+    def publish_and_receive(
+            self,
+            body: Any,
+            routing_key: str,
+            serializer: Optional[Callable] = None,
+            time_limit: int = 60,
+            headers: Optional[dict[str, Any]] = None
+    ) -> Any:
         """Publish message to a rabbit queue with the given routing key.
 
         :param body: The message to publish.
         :param routing_key: The routing key.
-        :param response_type: Type to map the response to, must be
-            annotated with @dataclass_json.
-        :param time_limit: Amount of time to wait for a response,
-            defaults to 10 seconds.
-        :param headers: To publish with.
+        :param serializer: Callable to serialize response bodies.
+        :param time_limit: Number of seconds to wait for a response.
+        :param headers: Headers to publish with.
         """
-        log.debug('Publishing key={} msg={}', routing_key, body)
+        log.debug('Publishing key=%s msg=%s', routing_key, body)
         mq_conn = BlockingConnection(self.config.params)
         channel: BlockingChannel = mq_conn.channel()
         response: Optional[bytes] = None
         corr_id = str(uuid.uuid4())
 
-        def _on_response(_channel: BlockingChannel,
-                         _method: Basic.Deliver,
-                         props: BasicProperties,
-                         resp_body: bytes) -> None:
+        def _on_response(
+                _channel: BlockingChannel,
+                _method: Basic.Deliver,
+                props: BasicProperties,
+                resp_body: bytes
+        ) -> None:
             nonlocal response
             if props.correlation_id == corr_id:
                 response = resp_body
@@ -86,29 +99,31 @@ class Rabbit:
                 correlation_id=corr_id,
                 headers=headers
             ),
-            body=_get_body(body)
+            body=body
         )
         mq_conn.process_data_events(time_limit=time_limit)
-        return response_type.from_json(response)
+        return serializer(response) if serializer else response
 
-    def listener(self, queues: list[str]) -> callable:
+    def listener(self, queues: list[str], serializer=None) -> Callable:
         """Decorate a function to make it a rabbit consumer.
 
         The decorated function can have as parameters, the message body,
         the method, the properties, or no parameters.
+        A new thread is made for each consumer.
 
         :param queues: The queues to consume from.
+        :param serializer: Callable to serialize message bodies.
         :return: Function decorated to be a rabbit consumer.
         """
         mq_conn = BlockingConnection(self.config.params)
         channel = mq_conn.channel()
 
-        def wrapper(fun: callable) -> None:
+        def wrapper(fun: Callable) -> None:
             original_name = fun.__name__
             for queue in queues:
                 fun.__name__ = f'listener_{queue}'
-                _listen(channel, self.config.exchange, queue, fun)
-                log.debug('{} subscribed to {}', original_name, queue)
+                _listen(channel, self.config.exchange, queue, fun, serializer)
+                log.info('%s subscribed to %s', original_name, queue)
             thread = threading.Thread(target=_consume, args=(channel,))
             thread.daemon = True
             thread.start()
@@ -116,21 +131,26 @@ class Rabbit:
         return wrapper
 
 
-def _listen(channel: BlockingChannel,
-            exchange: str,
-            queue: str,
-            fun: callable) -> None:
-    def callback(ch: BlockingChannel,
-                 method: Basic.Deliver,
-                 properties: BasicProperties,
-                 body: bytes) -> None:
-        log.debug('{} received {}', queue, body)
+def _listen(
+        channel: BlockingChannel,
+        exchange: str,
+        queue: str,
+        fun: Callable,
+        serializer: Optional[Callable] = None
+) -> None:
+    def callback(
+            ch: BlockingChannel,
+            method: Basic.Deliver,
+            properties: BasicProperties,
+            body: bytes
+    ) -> None:
+        log.debug('%s received %s', queue, body)
         sig = signature(fun)
-        kwargs: dict[str, any] = {}
+        kwargs: dict[str, Any] = {}
         for name, param in sig.parameters.items():
             if name == 'self':
                 kwargs[name] = None
-            if param.annotation == BlockingChannel:
+            elif param.annotation == BlockingChannel:
                 kwargs[name] = ch
             elif param.annotation == method:
                 kwargs[name] = method
@@ -144,14 +164,13 @@ def _listen(channel: BlockingChannel,
                 kwargs[name] = body
             elif param.annotation in {int, float}:
                 kwargs[name] = param.annotation(body)
+            elif 'from_json' in dir(param.annotation):
+                kwargs[name] = param.annotation.from_json(body)
             else:
                 try:
-                    kwargs[name] = param.annotation.from_dict(json.loads(body))
-                except AttributeError:
-                    try:
-                        kwargs[name] = json.loads(body)
-                    except json.decoder.JSONDecodeError:
-                        kwargs[name] = body
+                    kwargs[name] = json.loads(body)
+                except json.decoder.JSONDecodeError:
+                    kwargs[name] = body
         # noinspection PyBroadException
         try:
             resp = fun(**kwargs)
@@ -162,7 +181,7 @@ def _listen(channel: BlockingChannel,
                     properties=BasicProperties(
                         correlation_id=properties.correlation_id
                     ),
-                    body=_get_body(resp)
+                    body=serializer(resp) if serializer else resp
                 )
                 channel.basic_ack(delivery_tag=method.delivery_tag)
         except Exception as e:
@@ -182,12 +201,3 @@ def _consume(channel: BlockingChannel) -> None:
     except KeyboardInterrupt:
         channel.stop_consuming()
         channel.close()
-
-
-def _get_body(body: any) -> bytes:
-    if 'to_json' in dir(body):
-        return body.to_json().encode()
-    elif isinstance(body, bytes):
-        return body
-    else:
-        return str(body).encode()
